import os
import json

from multiprocessing.pool import ThreadPool
from subprocess import Popen, run, PIPE
from typing import List, Tuple

from .utils import tmpfile


class Clip(object):
    def ffprobe(self, entries, stream=None) -> dict:
        command = ['ffprobe', '-v', 'error', '-of', 'json',
                   '-show_entries', entries]

        if stream:
            command += ['-select_streams', stream]

        command += [self.path]

        proc = run(command, stdout=PIPE)
        return json.loads(proc.stdout)

    def __init__(self, path: str, container: str = 'wav', tmpfile = None):
        self.name = os.path.basename(path)
        self.path = path
        self._tmpfile = tmpfile

        self.container = container

        info = self.ffprobe('format=duration,start_time,format_name')['format']
        self.start = float(info.get('start_time', 0))

        duration = float(info.get('duration', 0))
        format_name = info.get('format_name', 'mpegts')

        if format_name == 'mov,mp4,m4a,3gp,3g2,mj2':
            self.end = duration
            self._duration = self.end - self.start
        else:
            self._duration = duration
            self.end = self.start + self.duration

        self.__duration = self._duration
        self.inpoint = self.start
        self.outpoint = self.end

        info = self.ffprobe('stream=height', 'v:0')['streams']
        if len(info) > 0:
            self.height = info[0].get('height', 0)
        else:
            self.height = 0

    def keyframes(self) -> Tuple[float, float, bool]:
        command = ['ffprobe',
                   '-v', 'error',
                   '-of', 'csv',
                   '-show_frames', 
                   '-select_streams', 'v:0',
                   '-skip_frame', 'nokey',
                   '-show_entries', 'frame=pts_time',
                   self.path]

        ff = Popen(command, stdout=PIPE)
        
        frames = []
        for i, line in enumerate(ff.stdout):
            frames.append(float(line.decode().split(',')[1]))

            if i >= 2: break

        ff.terminate()
        ff.wait()

        if len(frames) < 3:
            print(f'WARN: Clip {self.name} is too short to determine '
                  'frame monotonicity')
            return 0, 0, False

        offset = frames[0]
        step = frames[1] - offset
        monotonous = (frames[2] - offset - step * 2) == 0

        return offset, step, monotonous

    @property
    def duration(self):
        return self._duration

    @duration.setter
    def duration(self, new_value: float):
        if new_value <= self.__duration:
            self._duration = new_value

    def __del__(self):
        if self._tmpfile:
            os.unlink(self._tmpfile)

    def slice(self, start: float = 0, duration: float = None, chunks: int = 1,
              output_options: List[str] = []):
        """Split this Clip into one or multiple temporary Clips.

        By default splits only the audio track, outputting chunks
        in WAV format.
        """

        if not duration:
            duration = self.duration

        command = (f'ffmpeg -y -v error -ss {start}').split()
        command += ['-i', self.path]

        if start > self.duration:
            return []

        results = []
        for i in range(chunks):
            chunk_end = start + duration * (i + 1)
            if self.duration < chunk_end:
                duration -= chunk_end - self.duration

            if duration <= 0:  # nothing left
                break

            tmp_file_name = tmpfile()
            output = (f'-f {self.container} '
                      f'-ss {duration * i} '
                      f'-t {duration}').split()
            output += output_options
            output += [tmp_file_name]
            command += output
            results += [tmp_file_name]

        if run(command).returncode != 0:
            [os.unlink(chunk) for chunk in results]
            raise Exception('ffmpeg exited with non-zero code')

        return [Clip(chunk,
                     tmpfile=chunk,
                     container=self.container)
                for chunk in results]

    def slice_generator(self, duration: float,
                        start: float = None, end: float = None,
                        reverse: bool = False, **kwargs):
        pool = ThreadPool(1)
        kwargs['chunks'] = 1

        if not start:
            start = 0

        if not end:
            end = self.duration

        if not reverse:
            position = start
        else:
            position = end - duration

        async_result = None

        while (position < end) if not reverse else (position > start):
            result = None

            if async_result:
                result = position, async_result.get()[0]

                if not reverse:
                    position += duration
                else:
                    position -= duration

                    if position < start:
                        duration -= start - position
                        position = start

            async_result = pool.apply_async(self.slice, kwds=kwargs,
                                            args=(position, duration))

            if not result:
                continue

            try:
                yield result
            except GeneratorExit:
                break

        pool.close()



from typing import Any, Dict, Union, List
from retry_requests import retry
from datetime import datetime
from hashlib import sha1
from urllib.parse import urlparse
import dateutil.parser as dp

try:
    from streamlink import Streamlink
    from streamlink.exceptions import NoPluginError
except ImportError:
    Streamlink = None

    class NoPluginError(Exception):
        pass


# Source: https://raw.githubusercontent.com/TwitchRecover/TwitchRecover/main/domains.txt
# TODO: Fetch updated list if possible
VOD_DOMAINS = [
    'vod-metro.twitch.tv',
    'vod-pop-secure.twitch.tv',
    'vod-secure.twitch.tv',
    # 'd1g1f25tn8m2e6.cloudfront.net',
    # 'd1m7jfoe9zdc1j.cloudfront.net',
    # 'd1mhjrowxxagfy.cloudfront.net',
    # 'd1oca24q5dwo6d.cloudfront.net',
    # 'd1w2poirtb3as9.cloudfront.net',
    # 'd1xhnb4ptk05mw.cloudfront.net',
    # 'd1ymi26ma8va5x.cloudfront.net',
    # 'd2aba1wr3818hz.cloudfront.net',
    # 'd2dylwb3shzel1.cloudfront.net',
    # 'd2e2de1etea730.cloudfront.net',
    # 'd2nvs31859zcd8.cloudfront.net',
    # 'd2um2qdswy1tb0.cloudfront.net',
    # 'd2vjef5jvl6bfs.cloudfront.net',
    # 'd2xmjdvx03ij56.cloudfront.net',
    # 'd3fi1amfgojobc.cloudfront.net',
    # 'd36nr0u3xmc4mm.cloudfront.net',
    # 'd3aqoihi2n8ty8.cloudfront.net',
    # 'd3c27h4odz752x.cloudfront.net',
    # 'd3vd9lfkzbru3h.cloudfront.net',
    # 'd6d4ismr40iw.cloudfront.net',
    # 'd6tizftlrpuof.cloudfront.net',
    # 'ddacn6pr5v0tl.cloudfront.net',
    # 'dgeft87wbj63p.cloudfront.net',
    # 'dqrpb9wgowsf5.cloudfront.net',
    # 'ds0h3roq6wcgc.cloudfront.net',
    # 'dykkng5hnh52u.cloudfront.net',
]


# Source: https://github.com/TwitchRecover/TwitchRecover/blob/
# 48b32dccec752961b6402fff50eefcdc97ca27ff
# Source: https://github.com/tanersb/TwitchRecover/blob/
# bee8cc29fd44b00070c96c4c4c0d1b6ad811dcbd/recover.py#L14-L42
# /src/TwitchRecover.Core/Compute.java#L51
def vod_path(channel: str, stream_id: str,
             started_at: datetime, quality: str) -> str:
    base = f'{channel.lower()}_{stream_id}_{int(started_at.timestamp())}'
    hash = sha1(base.encode()).hexdigest()[:20]
    return f'/{hash}_{base}/{quality}/index-dvr.m3u8'


class TwitchException(Exception):
    pass


class ChannelNotFoundException(TwitchException):
    def __init__(self):
        super().__init__('Channel not found')


class StreamOfflineException(TwitchException):
    def __init__(self):
        super().__init__('Stream appears to be offline')


class VodException(TwitchException):
    pass


class VodNotFoundException(VodException):
    def __init__(self):
        super().__init__('VOD not found')


class GqlException(TwitchException):
    def __init__(self, text):
        super().__init__(text)


class VodTypeMismatchException(VodException):
    def __init__(self, expected, actual):
        super().__init__(f'Stream type is "{actual}" instead of "{expected}"')


class TwitchAPI:
    def __init__(self, headers: Dict[str, str] = {}):
        self.session = retry()

        if Streamlink:
            self.sl = Streamlink()
            self.session.headers = self.sl.http.headers
        else:
            self.sl = None

        self.session.headers['Client-ID'] = 'ue6666qo983tsx6so1t0vnawi233wa'
        self.session.headers.update(headers)

    def gql(self, query: str) -> dict:
        res = self.session.post('https://gql.twitch.tv/gql', json={'query': query})

        if res.status_code == 200:
            return res.json()
        else:
            raise GqlException(res.text)

    def get_headers(self) -> Dict[str, Union[str, bytes]]:
        return dict(self.session.headers)

    def get_stream(self, login: str) -> Dict[str, Any]:
        res = self.gql(f'''
            query {{
                user(login: "{login}") {{
                    stream {{
                        id
                        broadcaster {{
                            login
                        }}
                        createdAt
                        archiveVideo {{
                            id
                        }}
                        type
                    }}
                }}
            }}
        ''')

        user: Dict[str, Any] = res['data']['user']

        if user is None:
            raise ChannelNotFoundException

        stream: Dict[str, Any] = user['stream']

        if stream is None:
            raise StreamOfflineException
        
        return stream

    def get_active_vod(self, stream: Dict[str, Any], stream_type: str = 'live'):
        """Returns ID of VOD if stream is live."""

        if stream['type'] != stream_type:
            raise VodTypeMismatchException(stream_type, stream['type'])

        vod: Dict[str, Any] = stream['archiveVideo']

        if not vod:
            raise VodNotFoundException

        return vod['id']

    def get_vod_ids(self, login: str, first: int = 10) -> List[str]:
        res = self.gql(f'''
            query {{
                user(login: "{login}") {{
                    videos(first: {first}) {{
                        edges {{
                            node {{
                                id
                            }}
                        }}
                    }}
                }}
            }}
        ''')

        videos = res['data']['user']['videos']['edges']
        return [video['node']['id'] for video in videos]

    def vod_probe_domain(self, login: str) -> Union[str, None]:
        if not self.sl:
            return None

        try:
            prev_vod = self.get_vod_ids(login, first=1)[0]
        except IndexError:
            return None

        try:
            streams = self.sl.streams(f'twitch.tv/videos/{prev_vod}')
        except NoPluginError:
            return None

        stream = list(streams.values())[0]
        url = urlparse(stream.url)
        return url.hostname

    def vod_probe(self, stream: Dict[str, Any], quality: str = 'chunked') -> str:
        """Returns URL of VOD's playlist."""
        stream_id = stream['id']
        login = stream['broadcaster']['login']
        started_at = dp.parse(stream['createdAt'])

        # Try domain from previous VOD first
        predicted_domain = self.vod_probe_domain(login)

        if predicted_domain and predicted_domain not in VOD_DOMAINS:
            VOD_DOMAINS.append(predicted_domain)

        domains = sorted(VOD_DOMAINS, key=lambda x: x != predicted_domain)

        for domain in domains:
            path = vod_path(login, stream_id, started_at, quality)
            url = f'https://{domain}{path}'
            res = self.session.head(url, timeout=5)
            print(f'[{res.status_code}] {url}')

            if res.status_code == 200:
                return url

        raise VodNotFoundException

    def is_still_live(self, stream: Dict[str, Any]) -> bool:
        try:
            channel = stream['broadcaster']['login']
            return self.get_stream(channel)['id'] == stream['id']
        except StreamOfflineException:
            return False



"""Usage:
    twitch_utils record [options] [--header=<arg>]... <channel> [<vod>]

Parameters:
  channel   Name of the channel. Can be found in the URL: twitch.tv/<channel>
  vod       VOD of stream that is currently live. Can be found in the URL:
            twitch.tv/videos/<vod>

Options:
  --quality <value> Choose stream quality to save. Accepts the same values as
                    streamlink. To check available options use command
                    `streamlink twitch.tv/<channel>`. [default: best]
  --oauth <token>   Twitch OAuth token. You need to extract it from the site's
                    cookie named "auth-token". Equivalent to
                    header="Authorization=OAuth <token>".
  --header <value>  Add custom headers to all Twitch API calls (including
                    streamlink). Example: "X-Device-Id=value".
  -o <name>         Name of the output file. For more information see
                    `twitch_utils concat --help`. Defaults to `<vod>.ts`.
  -j <threads>      Number of concurrent downloads of live segments. [default: 4]
  -y, --force       Overwrite output file without confirmation.
  --no-concat       Only download all parts of the stream and ensure that
                    concatenation is possible.
  --debug           Forward output of streamlink and ffmpeg to stderr.
"""

import os
import sys
from itertools import count
from datetime import datetime
from dataclasses import dataclass, field
from typing import Callable, Dict, Union

from twitch_utils.hls import SimpleHLS

try:
    import streamlink
    from parse import compile
except ImportError:
    print('Error: You need to install tdh-twitch-utils[record] or '
          'tdh-twitch-utils[all] to use this feature.',
          file=sys.stderr)
    sys.exit(1)

from time import sleep
from threading import Lock, Thread, Event
from docopt import docopt
from subprocess import Popen, PIPE, DEVNULL

from .clip import Clip
from .concat import Timeline, MissingRangesError
from .twitch import TwitchAPI, VodException, VodNotFoundException


DEBUG = False


class Stream(object):
    PARSE_QUEUED = compile('{} Adding segment {segment:d} to queue{}')
    PARSE_COMPLETE = compile('{} Segment {segment:d} complete{}')
    PARSE_FAILED = compile('{} Download of segment {segment:d} failed{}')
    PARSE_DISCARDED = compile('{} Discarding segment {segment:d}{}')
    PARSE_SKIPPED = compile('{} Skipped segments {data} after playlist reload{}')

    def __init__(self, url: str,
                 quality: Union[str, None] = 'best',
                 threads: int = 1,
                 api: Union[TwitchAPI, None] = None,
                 live: bool = False):
        self.url = url
        self.quality = quality
        self.threads = threads
        self.api = api
        self.live = live

        self.result = -1
        self.started = Event()
        self._stream_url = None

    def copy(self):
        return Stream(self.url, self.quality, self.threads,
                      self.api)

    def _args(self) -> list:
        params: Dict[str, Union[str, int]] = {
            'stream-segment-attempts': 5,
            'stream-segment-threads': self.threads
        }

        args = ['-l', 'debug', '--no-config', '--twitch-disable-ads',
                '--twitch-low-latency']
        args += [f'--{key}={value}' for key, value in params.items()]

        if self.api:
            for key, value in self.api.get_headers().items():
                args += [f'--twitch-api-header={key}={value}']

        args.append(self.url)

        if self.quality:
            args.append(self.quality)

        args.append('-O')

        return args

    def download(self, dest: str) -> int:
        """Exit codes: 0 - success, 1 - should retry, 2 - should stop"""

        print(f'Downloading `{self.url}` into {dest}')
        exit_code = 0

        fo = open(dest, 'wb')

        sl_cmd = ['streamlink'] + self._args()
        sl_env = {**os.environ, 'PYTHONUNBUFFERED': '1'}
        sl_kwargs = {'stdout': PIPE,
                     'stderr': PIPE,
                     'text': True,
                     'env': sl_env}
        sl_proc = Popen(sl_cmd, **sl_kwargs)

        ff_cmd = ['ffmpeg', '-hide_banner',
                  '-i', '-',
                  '-c', 'copy', '-copyts',
                  '-f', 'mpegts', '-']
        ff_kwargs = {'stdin': sl_proc.stdout,
                     'stdout': fo,
                     'stderr': DEVNULL}
        ff_proc = Popen(ff_cmd, **ff_kwargs)

        sl_proc.stdout.close()

        expected, downloaded = [-1] * 2
        first_segment = True
        ts = 0

        while True:
            line = sl_proc.stderr.readline()

            if not line:
                break

            if DEBUG:
                print(line.rstrip(), file=sys.stderr)

            queued = Stream.PARSE_QUEUED.parse(line)
            complete = Stream.PARSE_COMPLETE.parse(line)
            failed = Stream.PARSE_FAILED.parse(line)
            discarded = Stream.PARSE_DISCARDED.parse(line)
            skipped = Stream.PARSE_SKIPPED.parse(line)

            if fo.tell() > 0 and ts > 0 and not self.started.is_set():
                # Log precise timings to leave some traces for manual
                # checks of recording's consistency
                # For example: If the following calculation
                #       (ts2 - ts1) - (inpoint2 - inpoint1)
                # is > 0, we can assume that Twitch has lost some segments
                # and the stream has become shorter by this amount.
                inpoint = Clip(dest).inpoint
                print(f'Clip {dest} started at {ts} with offset {inpoint}')
                self.started.set()

            if failed or (not first_segment and discarded):
                print(f'ERR: {line}')
                sl_proc.terminate()
                exit_code = 1
                break
            elif queued:
                segment = queued['segment']

                if queued['segment'] > expected:
                    expected = queued['segment']
            elif skipped:
                print(f'ERR: Skipped segments {skipped["data"]}')
                sl_proc.terminate()
                exit_code = 1
                break
            elif complete:
                segment = complete['segment']

                if self.live and first_segment:
                    # First write is delayed, log timestamp now
                    ts = datetime.now().timestamp()
                    first_segment = False

                if downloaded == -1 or downloaded + 1 == segment:
                    downloaded = segment
                else:
                    print(f'ERR: Skipped segment {downloaded + 1}')
                    sl_proc.terminate()
                    exit_code = 1
                    break
            elif 'No playable streams found' in line:
                print('WARN: Stream appears to be offline')
                sl_proc.terminate()
                exit_code = 2
                break
            elif 'Waiting for pre-roll ads' in line:
                print('Waiting for ads to finish...')

        sl_proc.wait()
        ff_proc.wait()

        fo.flush()
        fo.close()

        if downloaded < expected:
            print(f'ERR: Skipped segment {expected}')
            exit_code = 1

        return exit_code

    def _target_download(self, *args):
        self.result = None
        self.result = self.download(*args)

        # Unlock waiting threads if exit was early
        self.started.set()
        self.started.clear()

    def download_async(self, dest: str) -> Thread:
        self.started.clear()
        p = Thread(target=self._target_download, args=(dest,))
        p.start()
        return p

    def stream_url(self) -> str:
        if not self.live and self._stream_url:
            return self._stream_url

        sl_cmd = ['streamlink'] + self._args() + ['--stream-url']
        sl_env = {**os.environ, 'PYTHONUNBUFFERED': '1'}
        sl_kwargs = {'stdout': None,
                     'stderr': PIPE,
                     'text': True,
                     'env': sl_env}
        sl_proc = Popen(sl_cmd, **sl_kwargs)
        sl_proc.wait()

        url = sl_proc.stderr.readline().strip()
        self._stream_url = url
        return url

    def clip(self) -> Clip:
        return Clip(self.stream_url())


def generate_filename(vod_id, part):
    return f'{vod_id}.{part}.ts'


def create_timeline(vod_id, parts):
    clips = []

    for part in range(parts):
        filename = generate_filename(vod_id, part)

        try:
            clips.append(Clip(filename))
        except Exception:
            print(f'WARN: Clip {filename} is corrupted, ignoring...')

    return Timeline(clips)


class Counter:
    def __init__(self, value = 0):
        self._value = value
        self.lock = Lock()

    def set(self, value):
        with self.lock:
            self._value = value
            return self._value

    def inc(self):
        with self.lock:
            self._value += 1
            return self._value
    
    def dec(self):
        with self.lock:
            self._value -= 1
            return self._value

    @property
    def value(self):
        return self._value


@dataclass
class RecordingSession:
    vod: str
    api: Union[TwitchAPI, None] = None
    counter: Counter = field(default_factory=Counter)
    recording: Event = field(default_factory=Event)
    dirty: Event = field(default_factory=Event)

    def next_file(self) -> str:
        return generate_filename(self.vod, self.counter.inc() - 1)


class RecordThread(Thread):
    def __init__(self, session: RecordingSession,
                 channel_name: str,
                 quality: str = 'best', threads: int = 4,
                 is_online: Callable = lambda: True):
        super().__init__()

        self.session = session
        self.stream = Stream(f'https://twitch.tv/{channel_name}',
                             api=session.api,
                             quality=quality,
                             threads=threads,
                             live=True)
        self.is_online = is_online

    def run(self):
        result = -1

        while result != 0:
            filename = self.session.next_file()
            proc = self.stream.download_async(filename)

            # Wait for possible early exit
            self.stream.started.wait()

            if self.stream.result == 2:
                if self.session.counter.value == 1:
                    sys.exit(1)
                else:
                    break

            self.session.recording.set()
            self.session.dirty.set()

            proc.join()
            result = self.stream.result

            print(f'Finished download of live stream (exit code: {result})')

            if result != 0:
                print('Resuming in 60 seconds...')
                sleep(60)

                if not self.is_online():
                    print('Stream ended')
                    break

        self.session.recording.clear()
        self.session.dirty.set()


class RepairThread(Thread):
    def __init__(self, session: RecordingSession,
                 vod_url: Union[str, None] = None,
                 quality: str = 'best'):
        super().__init__()

        self.session = session

        if not vod_url:
            vod_url = f'https://twitch.tv/videos/{session.vod}'

        self.stream = Stream(vod_url, api=session.api, quality=quality)
        self.hls = None

    def first_vod(self):
        try:
            timeline = create_timeline(self.session.vod, self.session.counter.value)
            start = timeline.start
        except MissingRangesError as ex:
            start = ex.start

        result = None
        filename = self.session.next_file()

        while not result:
            if result is not None:
                print('WARN: Could not download VOD (exit code '
                        f'{result}), retrying in 60 seconds...')
                sleep(60)

            with open(filename, 'wb') as fo:
                self.hls.download(fo, end = start + 30)

            try:
                Clip(filename)
                result = True
            except Exception:
                result = False

        print(f'Finished download of VOD (exit code: {result})')

    @staticmethod
    def optimize_missing(ranges: list[tuple]) -> list[tuple]:
        ranges = list(ranges)
        ranges.sort(key=lambda x: x[0])

        optimized = []

        for i in range(len(ranges)):
            if i == 0:
                optimized.append(ranges[i])
                continue

            (a1, a2), (b1, b2) = optimized[-1], ranges[i]

            if (a2 - a1 + b2 - b1) > (b1 - a2):
                optimized[-1] = (a1, b2)
            else:
                optimized.append(ranges[i])

        return optimized

    def run(self):
        self.session.recording.wait()

        if not self.hls:
            self.hls = SimpleHLS(self.stream.stream_url())

        offset = self.hls.offset()

        missing_parts = None

        # Retry first VOD at least until it is readable
        if self.session.counter.value == 1:
            print('Starting download of VOD')
            self.first_vod()

        while self.session.recording.is_set():
            while missing_parts or self.session.dirty.is_set():
                self.session.dirty.clear()
                print('Testing the possibility of concatenation')

                try:
                    create_timeline(self.session.vod, self.session.counter.value)
                    print('Timeline is complete, good!')
                    missing_parts = None
                    break
                except MissingRangesError as ex:
                    if ex.start - offset > 1:
                        ex.ranges.append((offset, ex.start))

                    missing_parts = self.optimize_missing(ex.ranges)
                    print(f'WARN: {ex}')

                for (start, end) in missing_parts:
                    start_o = max(0, start - 30 - offset)
                    end_o = end + 30 - offset
                    print(f'Downloading segment {start}~{end}'
                          f'({start_o}~{end_o})')

                    filename = self.session.next_file()

                    with open(filename, 'wb') as fo:
                        self.hls.download(fo,
                                          start = start_o,
                                          end = end_o)

            if self.session.recording.is_set():
                self.session.dirty.wait()

        print('All parts are downloaded!')


def main(argv=None):
    args = docopt(__doc__, argv=argv)

    global DEBUG
    DEBUG = args['--debug']

    output = args['-o']

    if output == '-':
        print('ERR: This script does not support stdout as output.')
        sys.exit(1)

    channel = args['<channel>']

    api = None
    stream = None
    vod = args['<vod>']
    vod_url = None

    headers = {}

    for header in args['--header']:
        key, value = header.split('=', 1)
        headers[key] = value

    if args['--oauth']:
        headers['Authorization'] = 'OAuth ' + args['--oauth']

    api = TwitchAPI(headers)
    stream = api.get_stream(channel)

    try:
        if not vod:
            vod = api.get_active_vod(stream)
    except VodException:
        vod = stream['id']
        print(f'WARN: VOD is not listed, using stream ID {vod} as base name')

    session = RecordingSession(vod, api)

    for i in count():
        filename = generate_filename(vod, i)
        if not os.path.exists(filename):
            session.counter.set(i)
            break

    if (resumed := session.counter.value > 0):
        print('Found previous segments, resuming download')

    def is_online():
        if api and stream:
            try:
                return api.is_still_live(stream)
            except Exception:
                return True
        else:
            return True

    print('Starting record thread')
    record = RecordThread(session, channel, args['--quality'], args['-j'], is_online)
    record.start()

    if vod == stream['id']:
        if not resumed:
            session.recording.wait()

        print('Attempting to find the VOD playlist')

        clip = None

        for i in range(session.counter.value):
            try:
                clip = Clip(generate_filename(vod, i))
            except Exception:
                pass

        if not clip:
            print('ERR: No readable segments found, unable to get resolution')
            sys.exit(1)

        height = clip.height
        qualities = []
        if height >= 720:
            qualities += [f'{height}p60', f'{height}p30']
        else:
            qualities += [f'{height}']
        qualities += ['chunked']

        for i in qualities:
            try:
                vod_url = api.vod_probe(stream, i)
            except VodNotFoundException:
                pass

        if not vod_url:
            print(f'ERR: VOD not found')
            sys.exit(1)

    print('Starting repair thread')
    repair = RepairThread(session, vod_url, args['--quality'])
    repair.start()

    record.join()
    repair.join()

    parts = session.counter.value

    try:
        t = create_timeline(vod, parts)
    except MissingRangesError as ex:
        print(ex)
        print('ERR: Unable to concatenate segments!')
        sys.exit(1)

    if not args['--no-concat']:
        if not output:
            output = f'{vod}.ts'

        print(f'Writing stream recording to {output}')
        t.render(output, force=args['--force'])

        print('Cleaning up...')
        [os.unlink(generate_filename(vod, part)) for part in range(parts)]
    else:
        if not output:
            output = f'{vod}.mp4'

        files = ' '.join(generate_filename(vod, part) for part in range(parts))

        print('Use this command to concatenate parts into a full video:')
        print(f'> twitch_utils concat {files} -o {output}')


if __name__ == '__main__':
    main()



"""Usage: twitch_utils offset [options] [--] FILE1 FILE2

This script matches one chunk from FILE1 against all chunks of
FILE2, returning offset of FILE1 and cross-correlation score.

Matching is performed by cross-correlation of audio tracks.
Video, subtitles and metadata will be ignored.

Options:
  -s <t>, --start <t>       Skip <t> seconds at the beggining of FILE2. [default: 0]
  -e <t>, --end <t>         Stop matching at this offset of FILE2.
  -t <t>, --split <t>       Split FILE2 into chunks of this length. [default: 300]
  --template-start <t>      Template chunk will be cut from FILE1 starting at this offset. [default: 0]
  --template-duration <t>   Duration of template chunk. [default: 120]
  -r <frequency>            WAV sampling frequency (lower is faster but less accurate). [default: 1000]
  --reverse                 Start from the end of the video.

Exit conditions:
  --score-multiplier <N>    Stop computation if local maximum score is at least
                            N times bigger than last local minimum. [default: 8]

  --min-score <value>       Minimum cross-correlation score to be treated as potential match.
                            This option is useful if input files have no collisions at all.
                            In this case both offset and score will be 0.
  --max-score <value>       Stop computation if cross-correlation score exceeds this value.

  WARNING: cross-correlation score depends on many factors such as segment
  length, audio sampling frequency and volume of the audio track. Be careful
  when using absolute values.

Output parameters:
  --round                   Output integer instead of float
  --score                   Output cross-correlation score along with offset

Usage examples:

  - Two video files

  offset.py template.mp4 long_video.mp4

  - Match beginning of YouTube video with local file

  offset.py $(youtube-dl -gf best VIDEO_ID) long_video.mp4
"""

import sys

try:
    import parselmouth as pm
except ImportError:
    print('Error: You need to install tdh-twitch-utils[offset] or '
          'tdh-twitch-utils[all] to use this feature.',
          file=sys.stderr)
    sys.exit(1)

from typing import Tuple
from docopt import docopt

from .clip import Clip


def offset(template: Clip, video: Clip) -> Tuple[float, float]:
    """Find position of this Clip in another Clip (may be negative).

    Returns two values: offset in seconds and cross-correlation score.
    """
    s1 = pm.Sound(template.path).convert_to_mono()
    s2 = pm.Sound(video.path).convert_to_mono()
    cc = s1.cross_correlate(s2, pm.AmplitudeScaling.SUM)
    score = cc.values.max()
    frame = cc.values.argmax()
    offset = cc.frame_number_to_time(frame)
    return offset, score


def find_offset(c1: Clip, c2: Clip,
                start: float = 0, end: float = None, reverse: bool = False,
                chunk_size: float = 300,
                ar: int = 500,
                min_score: float = None,
                max_score: float = None,
                score_multiplier: float = 8) -> Tuple[float, float]:

    last_best_offset, last_best_score = 0, 0
    last_worst_score = 0
    best_offset, best_score = 0, 0
    prev_score = 0

    print(f'pos | offset | score | mul', file=sys.stderr)

    c1 = c1.slice(output_options=['-vn', '-ar', str(ar)])[0]

    for position, chunk in c2.slice_generator(
            chunk_size, start, end, reverse,
            output_options=['-vn', '-ar', str(ar)]):
        new_offset, new_score = offset(c1, chunk)

        delta = new_score - prev_score
        prev_score = new_score

        if new_score > best_score:
            best_score = new_score
            best_offset = position + new_offset

        if delta > 0:
            last_best_score = new_score
            last_best_offset = position + new_offset
        else:
            last_worst_score = new_score

        if last_worst_score != 0:
            cur_multiplier = round(last_best_score / last_worst_score, 2)
        else:
            cur_multiplier = 'N/A'

        print(f'{position} | {round(new_offset, 2)} | '
              f'{round(new_score, 2)} | '
              f'{cur_multiplier}',
              file=sys.stderr)

        if max_score is not None and new_score >= max_score:
            return new_offset, new_score

        if last_worst_score > 0 and last_best_score > 0:
            if last_worst_score * score_multiplier < last_best_score:
                if min_score is not None and last_best_score < min_score:
                    continue
                return last_best_offset, last_best_score

    if min_score is None or best_score >= min_score:
        return best_offset, best_score
    else:
        return 0, 0


def main(argv=None):
    args = docopt(__doc__, argv=argv)

    def get_arg(key, default_value, func=lambda x: x):
        if key in args and args[key] is not None:
            return func(args[key])
        else:
            return default_value

    template_start = float(args['--template-start'])
    template_duration = float(args['--template-duration'])
    ar = int(args['-r'])

    c1 = Clip(args['FILE1'])

    if c1.duration < template_start or template_duration <= 0:
        raise Exception('Template is empty (check start offset and duration)')

    c1 = c1.slice(template_start, template_duration + template_start)[0]

    c2 = Clip(args['FILE2'])

    kwargs = {
        'start': float(args['--start']),
        'end': get_arg('--end', None, float),
        'chunk_size': float(args['--split']),
        'ar': ar,
        'min_score': get_arg('--min-score', None, float),
        'max_score': get_arg('--max-score', None, float),
        'score_multiplier': float(args['--score-multiplier']),
        'reverse': args['--reverse']
    }

    offset, score = find_offset(c1, c2, **kwargs)

    if offset == 0 and score == 0:
        raise Exception('Videos are not correlated or min-score is too high')

    offset -= template_start

    if args['--round']:
        offset = round(offset)
        score = round(score)

    if args['--score']:
        print(f'Offset: {offset}')
        print(f'Score: {score}')
    else:
        print(offset)


if __name__ == '__main__':
    main()



"""Usage:
    twitch_utils vod [options] (-c <channel> | -v <vod>) [-u <url>]

Parameters:
  output        Name of the output file. For more information see
                `twitch_utils concat --help`. Defaults to `<vod>.ts`.

Options:
  -u <url>      Force HLS playlist URL (bypass quality selection).
  -q <value>    Choose stream quality to save. To check available options
                use -Q. Chunked == source. [default: chunked]
  -Q            Displays resolutions available in the HLS playlist, then exits.
                Works only if the VOD is listed.
  -o <output>   Name of the output file. For more information see
                `twitch_utils concat --help`. Defaults to `<vod>.ts`.
  -y, --force   Overwrite output file without confirmation.
  --no-concat   Only download all parts of the stream and ensure that
                concatenation is possible.
"""

import os
import re
import sys
from itertools import count
from docopt import docopt
from typing import Dict

from .concat import MissingRangesError
from .record import Stream, RecordingSession, create_timeline, generate_filename
from .twitch import TwitchAPI, VodNotFoundException
from .hls import SimpleHLS


def parse_usher(res: str) -> Dict[str, str]:
    streams = dict()
    quality = None

    p = re.compile('.*VIDEO="(.*?)".*')

    for line in res.split('\n'):
        if line.startswith('#EXT-X-STREAM-INF:'):
            m = p.match(line)

            if m:
                quality = m.group(1)

        if not line.startswith('#') and quality:
            streams[quality] = line
            quality = None

    if 'chunked' not in streams:
        first_key = list(streams.keys())[0]
        streams['chunked'] = streams[first_key].replace(first_key, 'chunked')

    return streams


def resolve_playlist(args, api: TwitchAPI):
    url = args['-u']
    main_url = None

    if (vod := args['<vod>']):
        if url:
            return vod, url

        vod_obj = Stream(f'https://twitch.tv/videos/{vod}', None)
        main_url = vod_obj.stream_url()

    if (channel := args['<channel>']):
        stream = api.get_stream(channel)

        if not stream:
            print(f'ERR: Stream offline')
            sys.exit(1)

        try:
            vod = api.get_active_vod(stream)

            if url:
                return vod, url

            vod_obj = Stream(f'https://twitch.tv/videos/{vod}', None)
            main_url = vod_obj.stream_url()
        except VodNotFoundException:
            print('VOD is not listed, attempting to find the playlist')
            url = api.vod_probe(stream, args['-q'])
            vod = str(stream['id'])
            print(f'VOD found! Using stream ID {vod} as base name')
            return vod, url

    if main_url:
        res = api.session.get(main_url)

        if res.status_code != 200:
            print('ERR: Unable to read main playlist')
            sys.exit(1)

        streams = parse_usher(res.content.decode())

        if args['-Q']:
            print('Available streams: ' + ', '.join(streams.keys()))
            sys.exit(0)

        if args['-q'] not in streams:
            print(f'ERR: Quality {args["-q"]} is not available')
            sys.exit(1)

        url = streams[args['-q']]
    elif args['-Q']:
        print('ERR: Unable to find master playlist, -Q is not available')
        sys.exit(1)

    if not url:
        print(f'ERR: Unable to resolve URL of HLS playlist')

    return vod, url


def main(argv=None):
    args = docopt(__doc__, argv=argv)

    api = TwitchAPI()
    vod, url = resolve_playlist(args, api)
    session = RecordingSession(vod, api)
    hls = SimpleHLS(url)

    for i in count():
        filename = generate_filename(vod, i)
        if not os.path.exists(filename):
            session.counter.set(i)
            break
    
    if session.counter.value > 0:
        print('Found previous segments, resuming download')

    result = False

    while not result:
        start, end = 0, None

        if session.counter.value > 0:
            try:
                timeline = create_timeline(vod, session.counter.value)
                start = timeline.end - 120
                end = None
            except MissingRangesError as ex:
                start, end = ex.ranges[0]

        filename = session.next_file()
        print(f'Downloading {start}~{end} into {filename}')

        with open(filename, 'wb') as fo:
            result = hls.download(fo, start=start, end=end)

            if end:
                result = False

    output = args['-o']
    parts = session.counter.value

    if not output:
        output = f'{vod}.ts'

    try:
        t = create_timeline(vod, parts)
    except MissingRangesError as ex:
        print(ex)
        print('ERR: Unable to concatenate segments!')
        sys.exit(1)

    if parts == 1 and output.endswith('.ts') and not args['--no-concat']:
        print('Skipping concatenation (not required)')
        os.rename(generate_filename(vod, 0), output)
    elif not args['--no-concat']:
        print(f'Writing stream recording to {output}')
        t.render(output, force=args['--force'])

        print('Cleaning up...')
        [os.unlink(generate_filename(vod, part)) for part in range(parts)]
    else:
        files = ' '.join(generate_filename(vod, part) for part in range(parts))

        print('Use this command to concatenate parts into a full video:')
        print(f'> twitch_utils concat {files} -o {output}')


if __name__ == '__main__':
    main()


import os
from dataclasses import dataclass
from retry_requests import retry
from typing import List, Union, BinaryIO
from time import time, sleep
from subprocess import DEVNULL, Popen, PIPE

from .utils import tmpfile
from .clip import Clip

@dataclass
class HLSSegment:
    duration: float = 0
    name: str = '0.ts'
    init: bool = False
    final: bool = False


class SimpleHLS:
    def __init__(self, playlist: str) -> None:
        self.base_url = '/'.join(playlist.split('/')[:-1]) + '/'
        self.playlist = playlist.split('/')[-1]

        self.session = retry()
        self.segments: List[HLSSegment] = []
        self.last_reload = 0.0
        self.target_duration = 10.0
        self.reload()
    
    def reload(self):
        if len(self.segments) > 0:
            if time() < self.last_reload + self.target_duration / 2:
                sleep(self.target_duration / 2)

        self.last_reload = time()

        res = self.session.get(self.base_url + self.playlist)

        if res.status_code != 200:
            return

        segments: List[HLSSegment] = []
        segment = None

        for line in res.content.decode().splitlines():
            if line.startswith('#EXT-X-TARGETDURATION'):
                self.target_duration = float(line.split(':')[1])

            if line.startswith('#EXT-X-MAP:URI='):
                segment = HLSSegment()
                segment.name = line.split('=')[1].strip('"')
                segment.duration = 0
                segment.init = True

            if line.startswith('#EXTINF:'):
                if segment:
                    segments.append(segment)

                segment = HLSSegment()
                segment.duration = float(line[8:].split(',')[0])

            if not segment:
                continue

            if not line.startswith('#'):
                segment.name = line

            if line.startswith('#EXT-X-ENDLIST'):
                segment.final = True
                segments.append(segment)
                break
        
        self.segments = segments

    def iterate(self, start: float = 0, end: Union[float, None] = None):
        i = 0
        offset = 0

        while not end or offset < end:
            while len(self.segments) < i + 1:
                self.reload()

                if self.segments[i - 1].final:
                    raise StopIteration

            segment = self.segments[i]
            i += 1

            if segment.init:
                yield segment
                continue

            offset += segment.duration

            if offset < start:
                continue

            yield segment

            if segment.final:
                break

    def download(self, fo: BinaryIO, start: float = 0, end: Union[float, None] = None) -> bool:
        result = True

        ff_cmd = ['ffmpeg', '-hide_banner',
                  '-i', '-',
                  '-c', 'copy', '-copyts',
                  '-f', 'mpegts', '-']
        ff_kwargs = {'stdin': PIPE,
                     'stdout': fo,
                     'stderr': DEVNULL}
        ff_proc = Popen(ff_cmd, **ff_kwargs)

        for segment in self.iterate(start, end):
            res = self.session.get(self.base_url + segment.name, stream=True)

            if res.status_code != 200:
                print(f'Failed to download segment {segment.name} '
                      f'(code: {res.status_code})')
                result = False
                break

            for chunk in res.iter_content(4096):
                ff_proc.stdin.write(chunk)

        ff_proc.stdin.flush()
        ff_proc.stdin.close()
        ff_proc.wait()

        return result

    def offset(self):
        clip_file = tmpfile('ts')

        with open(clip_file, 'wb') as fo:
            self.download(fo, end = 30)

        clip = Clip(clip_file)
        offset = clip.start

        os.unlink(clip_file)

        return offset


if __name__ == '__main__':
    import sys

    with open(sys.argv[2], 'wb') as fo:
        hls = SimpleHLS(sys.argv[1])
        hls.download(fo)



"""Usage: twitch_utils concat [options] <input>... (-o <name> | --output <name>)

This script concatenates two or more MPEG-TS files deleting
overlapping segments in the process. Re-encoding is not required.

Clips are aligned by start_time field. For example, Twitch
streams and their corresponding VODs share the same timeline.
It means that you can record stream, download its beginning
later and concatenate them by using this script.

Supported output formats:
  * mp4 (*.mp4)
  * mpegts (*.ts, -)
  * flv (*.flv, -)
  * txt (*.txt, -), map file for ffmpeg's concat demuxer
  * edl (*.edl, -), EDL playlist for MPV (preview without concatenation)
  * edl_uri (-), inline EDL URI for MPV

Options:
  -y, --force                     Overwrite output file without confirmation.
  -f <format>, --format=<format>  Force output pipe format. Has no effect if
                                  output file name is specified. [default: mpegts]
  -o <name>, --output=<name>      Name of the output file. Use '-' to output
                                  directly to stdout.
  --min-overlap=<t>               Minimal safe overlap of clips in seconds
                                  [default: 10].

MP4 options:
  --faststart   Move moov atom to the front of the file. Requires second pass
                that makes concatenation much slower.
"""

import os
import sys
from typing import List

from docopt import docopt
from subprocess import run, PIPE

from .clip import Clip
from .utils import tmpfile


class MissingRangeError(Exception):
    def __init__(self, start, end):
        super().__init__(f'Range {start}~{end} is missing')
        self.start = start
        self.end = end
        self.range = (start, end)


class MissingRangesError(Exception):
    def __init__(self, ranges, start, end):
        super().__init__('Timeline incomplete, missing: ' +
                         ', '.join(f'{r[0]}~{r[1]}' for r in ranges))
        self.ranges = ranges
        self.start = start
        self.end = end


class Timeline(list):
    @staticmethod
    def find_clip(clips: list, pos: float) -> list[Clip]:
        end = None
        found = []

        for clip in clips:
            if clip.start <= pos and clip.end > pos:
                found.append(clip)

            if clip.start > pos:
                if not end or end > clip.start:
                    end = clip.start

        if len(found) == 0:
            raise MissingRangeError(pos, end)

        return found

    @staticmethod
    def overlap(a: Clip, b: Clip) -> tuple[float, float]:
        #                outpoint   end
        # a ================|--------|
        #          |--------|=============== b
        #        start   inpoint
        overlap = a.end - b.start
        middle = b.start + overlap / 2

        return overlap, middle

    def __init__(self, clips: list, min_overlap: float = 10):
        clips.sort(key=lambda k: k.start)

        self.start = min([clip.start for clip in clips])
        self.end = max([clip.end for clip in clips])

        pos = self.start
        offset, step, monotonous = clips[0].keyframes()
        missing = []

        while pos < self.end:
            try:
                candidates = self.find_clip(clips, pos)
            except MissingRangeError as ex:
                missing.append(ex.range)
                pos = ex.end
                continue

            if len(self) == 0:
                clip = max(candidates, key=lambda c: c.duration)
                self.append(clip)
                pos = clip.end
                continue

            a = self[-1]
            overlaps = list(map(lambda x: (x, *self.overlap(a, x)), candidates))

            max_overlap = max(overlaps, key=lambda x: x[1])[1]

            if max_overlap >= min_overlap:
                overlaps = filter(lambda x: x[1] >= min_overlap, overlaps)
                b, overlap, middle = min(overlaps, key=lambda x: x[1])
            else:
                b, overlap, middle = max(overlaps, key=lambda x: x[1])

            if monotonous:
                # Cut by keyframe closest to the middle
                frame = (middle - offset) // step
                middle = frame * step + offset

            a.outpoint = b.inpoint = middle

            self.append(b)
            pos = b.end

            if overlap >= 0 and overlap < min_overlap:
                missing.append((b.start, a.end))

        if len(missing) > 0:
            raise MissingRangesError(missing, self.start, self.end)

    def ffconcat_map(self) -> str:
        return '\n'.join([
            f"file '{c.path}'\ninpoint {c.inpoint}\noutpoint {c.outpoint}\n"
            for c in self
        ])

    def edl_parts(self) -> List[str]:
        return [
            f"%{len(c.path)}%{c.path},{c.inpoint},{c.outpoint - c.inpoint}"
            for c in self
        ]

    def edl_uri(self) -> str:
        return 'edl://' + ';'.join(self.edl_parts())

    def edl_map(self) -> str:
        return '# mpv EDL v0\n' + '\n'.join(self.edl_parts())

    def render(self, path: str = 'full.mp4', container: str = 'mp4',
               mp4_faststart: bool = False, force: bool = False) -> int:
        if path.endswith('.edl') or path == '-' and container == 'edl':
            if path == '-':
                print(self.edl_map())
            else:
                with open(path, 'w') as fo:
                    fo.write(self.edl_map())
                    fo.flush()
            return 0

        if path == '-' and container == 'edl_uri':
            print(self.edl_uri())
            return 0

        concat_map = self.ffconcat_map()

        if path.endswith('.txt') or path == '-' and container == 'txt':
            if path == '-':
                print(concat_map)
            else:
                with open(path, 'w') as fo:
                    fo.write(concat_map)
                    fo.flush()
            return 0

        print(concat_map, file=sys.stderr)

        map_file_name = tmpfile('txt', '.')
        map_file = open(map_file_name, 'w')
        map_file.write(concat_map)
        map_file.flush()
        map_file.close()

        command = ['ffmpeg']

        if force:
            command += ['-y']

        if path.endswith('.ts') or path == '-' and container == 'mpegts':
            command += ['-copyts']

        command += ['-f', 'concat', '-safe', '0', '-hide_banner',
                    '-i', map_file_name, '-c', 'copy']

        if path.endswith('.ts') or path == '-' and container == 'mpegts':
            command += ['-muxdelay', '0']
            if path == '-':
                command += ['-f', 'mpegts']
        elif path.endswith('.flv') or path == '-' and container == 'flv':
            command += ['-bsf:a', 'aac_adtstoasc']
            if path == '-':
                command += ['-f', 'flv']
        elif path.endswith('.mp4'):
            command += ['-fflags', '+genpts', '-async', '1']
            if mp4_faststart:
                command += ['-movflags', 'faststart']

        command += [path]

        p = run(command)

        os.unlink(map_file_name)

        return p.returncode


def main(argv=None):
    args = docopt(__doc__, argv=argv)

    clips = []

    for path in args['<input>']:
        try:
            clips.append(Clip(path))
        except Exception:
            print(f'WARN: Clip {path} is corrupted, ignoring...',
                  file=sys.stderr)

    try:
        timeline = Timeline(clips, min_overlap=float(args['--min-overlap']))
    except MissingRangesError as ex:
        print(f'ERROR: {ex}', file=sys.stderr)
        sys.exit(1)

    sys.exit(timeline.render(args['--output'],
                             container=args['--format'],
                             mp4_faststart=args['--faststart'],
                             force=args['--force']))


if __name__ == '__main__':
    main()



"""Usage: twitch_utils <command> [<args>...]

Commands:
  concat  Concatenate multiple MPEG-TS files
  offset  Calculate offset between two audio (or video) files
  mute    Remove background music from video
  record  Record live Twitch stream from the beginning
"""

from docopt import docopt


def main(argv=None):
    args = docopt(__doc__, argv=argv, options_first=True)

    argv = [args['<command>']] + args['<args>']

    if args['<command>'] == 'concat':
        from .concat import main as concat_main
        concat_main(argv)
    elif args['<command>'] == 'offset':
        from .offset import main as offset_main
        offset_main(argv)
    elif args['<command>'] == 'mute':
        from .mute import main as mute_main
        mute_main(argv)
    elif args['<command>'] == 'record':
        from .record import main as record_main
        record_main(argv)
    elif args['<command>'] == 'vod':
        from .vod import main as vod_main
        vod_main(argv)


if __name__ == '__main__':
    main()



#!/usr/bin/env python3

import itertools
from setuptools import setup, find_packages


with open('README.md', 'r') as fi:
    long_description = fi.read()


EXTRAS = {
    'record': [
        'streamlink>=7.4.0',
        'parse>=1.19.0'
    ],
    'offset': [
        'praat-parselmouth>=0.4'
    ],
    'mute': [
        'spleeter>=2.3.0'
    ]
}

EXTRAS['all'] = list(itertools.chain.from_iterable(EXTRAS.values()))


setup(
    name='tdh-twitch-utils',
    version='1.8.2',

    author='Dmitry Karikh',
    author_email='the.dr.hax@gmail.com',

    description='Record, concatenate and synchronize Twitch live streams',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/TheDrHax/Twitch-Utils',

    install_requires=[
        'requests',
        'retry-requests',
        'python-dateutil',
        'docopt>=0.6.2',
    ],

    extras_require=EXTRAS,

    packages=find_packages(),

    entry_points='''
    [console_scripts]
    twitch_utils=twitch_utils:main
    ''',

    classifiers=[
        'Environment :: Console',
        'Programming Language :: Python :: 3.6',
        'Intended Audience :: End Users/Desktop',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Topic :: Multimedia :: Sound/Audio :: Analysis',
        'Topic :: Multimedia :: Video :: Conversion',
        'Topic :: Utilities'
    ]
)



"""Usage: twitch_utils mute [options] <input> <range>... -o <output>

This script attempts to separate streamer's voice from background music by
using Spleeter. Only specified time ranges are affected. Output contains the
same video, but without music in these parts.

The main purpose of this script is to remove automated Content-ID claims from
the video on YouTube. YouTube displays exact time ranges for every claim.

Timestamp format: [[HH:]MM:]SS[.MMM]
Time range format: START~END

Options:
  -o <file>    Name of the output file.
  --inverse    Remove voice instead of music.
  --pass <n>   Number of passes for each time range. [default: 1]
  --chunk <s>  Split ranges into smaller chunks (in seconds). [default: 300]
"""

import os
import sys

from docopt import docopt
from subprocess import run

try:
    from spleeter.separator import Separator
    from spleeter.audio.adapter import AudioAdapter
except ImportError:
    print('Error: You need to install tdh-twitch-utils[mute] or '
          'tdh-twitch-utils[all] to use this feature.',
          file=sys.stderr)
    sys.exit(1)

from .clip import Clip
from .utils import tmpfile


def ptime(t: str) -> float:
    parts = list(map(float, t.split(':')))[::-1]
    return sum(part * 60**i for i, part in enumerate(parts))


def main(argv=None):
    args = docopt(__doc__, argv=argv)

    fi = Clip(args['<input>'])
    fo = args['-o']
    ranges = list(tuple(ptime(t) for t in range.split('~'))
                  for range in args['<range>'])

    loader = AudioAdapter.default()
    sample_rate = 44100
    chunk_size = float(args['--chunk'])
    separator = Separator('spleeter:2stems')

    new_ranges = []
    for r in ranges:
        start, end = r
        
        while end - start > chunk_size:
            new_ranges.append((start, start + chunk_size))
            start += chunk_size
        
        new_ranges.append((start, end))

    ranges = new_ranges
    segments = {}

    for start, end in ranges:
        print(f'Processing range {start}-{end}...')

        options = ['-vn', '-r', str(sample_rate), '-f', 'wav']
        clip = fi.slice(start, end - start, output_options=options)[0]

        for i in range(int(args['--pass'])):
            waveform, _ = loader.load(clip.path, sample_rate=sample_rate)
            prediction = separator.separate(waveform)

            output = tmpfile('wav')

            target = 'accompaniment' if args['--inverse'] else 'vocals'
            loader.save(output, prediction[target], sample_rate)

            clip = Clip(output, tmpfile=output)

        segments[start] = clip

    print('Writing output file...')

    # Mute ranges in the original audio track
    # asetnsamples is required, source: https://superuser.com/a/1230890
    filters = '[0:a]asetnsamples=8192,'
    filters += ','.join(f"volume=0:enable='between(t,{start},{end})'"
                        for start, end in ranges)
    filters += '[main]'

    # Delay processed segments
    for i, (start, end) in enumerate(ranges):
        delay = int(start * 1000)
        filters += f';[{i+1}]'
        filters += 'asetnsamples=8192'
        filters += f',adelay={delay}|{delay},apad[delay{i+1}]'

    # Mix muted original track and all processed segments
    filters += ';[main]'
    for i, (start, end) in enumerate(ranges):
        filters += f'[delay{i+1}]'
    filters += f'amix=inputs={len(ranges) + 1}:duration=first'
    filters += f',volume={len(ranges) + 1}'

    filters += '[audio]'

    command = ['ffmpeg', '-i', fi.path]

    for start, segment in segments.items():
        command += ['-i', segment.path]

    # Copy codecs from the original video
    ainfo = fi.ffprobe('stream=codec_name,bit_rate', 'a')['streams'][0]
    command += ['-c:v', 'copy',
                '-c:a', ainfo['codec_name'],
                '-b:a', ainfo['bit_rate'],
                '-strict', '-2']

    command += ['-filter_complex', filters,
                '-map', '0:v', '-map', '[audio]', fo]

    if run(command).returncode != 0:
        if os.path.exists(fo):
            os.unlink(fo)
        raise Exception('ffmpeg exited with non-zero code')


if __name__ == '__main__':
    main()
